@@ -1,38 +1,22 @@
-/* pam_deny module */
-
 /*
- * $Id: pam_deny.c,v 1.4 2005/12/14 09:58:32 kukuk Exp $
+ * pam_deny module
  *
  * Written by Andrew Morgan <morgan@parc.power.net> 1996/3/11
- *
- */
-
-/*
- * here, we make definitions for the externally accessible functions
- * in this file (these definitions are required for static modules
- * but strongly encouraged generally) they are used to instruct the
- * modules include file to define their prototypes.
  */
 
 #include "config.h"
-
-#define PAM_SM_AUTH
-#define PAM_SM_ACCOUNT
-#define PAM_SM_SESSION
-#define PAM_SM_PASSWORD
-
 #include <security/pam_modules.h>
 
 /* --- authentication management functions --- */
 
-PAM_EXTERN int
+int
 pam_sm_authenticate(pam_handle_t *pamh UNUSED, int flags UNUSED,
 		    int argc UNUSED, const char **argv UNUSED)
 {
      return PAM_AUTH_ERR;
 }
 
-PAM_EXTERN int
+int
 pam_sm_setcred(pam_handle_t *pamh UNUSED, int flags UNUSED,
 	       int argc UNUSED, const char **argv UNUSED)
 {
@@ -41,7 +25,7 @@ pam_sm_setcred(pam_handle_t *pamh UNUSED, int flags UNUSED,
 
 /* --- account management functions --- */
 
-PAM_EXTERN int
+int
 pam_sm_acct_mgmt(pam_handle_t *pamh UNUSED, int flags UNUSED,
 		 int argc UNUSED, const char **argv UNUSED)
 {
@@ -50,7 +34,7 @@ pam_sm_acct_mgmt(pam_handle_t *pamh UNUSED, int flags UNUSED,
 
 /* --- password management --- */
 
-PAM_EXTERN int
+int
 pam_sm_chauthtok(pam_handle_t *pamh UNUSED, int flags UNUSED,
 		 int argc UNUSED, const char **argv UNUSED)
 {
@@ -59,14 +43,14 @@ pam_sm_chauthtok(pam_handle_t *pamh UNUSED, int flags UNUSED,
 
 /* --- session management --- */
 
-PAM_EXTERN int
+int
 pam_sm_open_session(pam_handle_t *pamh UNUSED, int flags UNUSED,
 		    int argc UNUSED, const char **argv UNUSED)
 {
     return PAM_SESSION_ERR;
 }
 
-PAM_EXTERN int
+int
 pam_sm_close_session(pam_handle_t *pamh UNUSED, int flags UNUSED,
 		     int argc UNUSED, const char **argv UNUSED)
 {
@@ -74,16 +58,3 @@ pam_sm_close_session(pam_handle_t *pamh UNUSED, int flags UNUSED,
 }
 
 /* end of module definition */
-
-/* static module data */
-#ifdef PAM_STATIC
-struct pam_module _pam_deny_modstruct = {
-    "pam_deny",
-    pam_sm_authenticate,
-    pam_sm_setcred,
-    pam_sm_acct_mgmt,
-    pam_sm_open_session,
-    pam_sm_close_session,
-    pam_sm_chauthtok
-};
-#endif
