@@ -1,5 +1,5 @@
 /*
- * $Id: pam_filter.c,v 1.14 2009/04/03 00:36:25 ldv Exp $
+ * pam_filter module
  *
  * written by Andrew Morgan <morgan@transmeta.com> with much help from
  * Richard Stevens' UNIX Network Programming book.
@@ -25,11 +25,6 @@
 
 #include <signal.h>
 
-#define PAM_SM_AUTH
-#define PAM_SM_ACCOUNT
-#define PAM_SM_SESSION
-#define PAM_SM_PASSWORD
-
 #include <security/pam_modules.h>
 #include <security/pam_ext.h>
 #include "pam_filter.h"
@@ -78,13 +73,13 @@ static int process_args(pam_handle_t *pamh
 	} else if (strcmp("run1",*argv) == 0) {
 	    ctrl |= FILTER_RUN1;
 	    if (argc <= 0) {
-		pam_syslog(pamh, LOG_ALERT, "no run filter supplied");
+		pam_syslog(pamh, LOG_ERR, "no run filter supplied");
 	    } else
 		break;
 	} else if (strcmp("run2",*argv) == 0) {
 	    ctrl |= FILTER_RUN2;
 	    if (argc <= 0) {
-		pam_syslog(pamh, LOG_ALERT, "no run filter supplied");
+		pam_syslog(pamh, LOG_ERR, "no run filter supplied");
 	    } else
 		break;
 	} else {
@@ -114,38 +109,37 @@ static int process_args(pam_handle_t *pamh
 	    return -1;
 	}
 
-	for (size=i=0; i<argc; ++i) {
-	    size += strlen(argv[i])+1;
-	}
-
 	/* the "ARGS" variable */
 
-#define ARGS_OFFSET    5                          /*  strlen('ARGS=');  */
 #define ARGS_NAME      "ARGS="
+#define ARGS_OFFSET    (sizeof(ARGS_NAME) - 1)
 
-	size += ARGS_OFFSET;
+	size = sizeof(ARGS_NAME);
 
-	levp[0] = (char *) malloc(size);
+	for (i=0; i<argc; ++i) {
+	    size += strlen(argv[i]) + (i != 0);
+	}
+
+	levp[0] = malloc(size);
 	if (levp[0] == NULL) {
 	    pam_syslog(pamh, LOG_CRIT, "no memory for filter arguments");
-	    if (levp) {
-		free(levp);
-	    }
+	    free(levp);
 	    return -1;
 	}
 
-	strncpy(levp[0],ARGS_NAME,ARGS_OFFSET);
-	for (i=0,size=ARGS_OFFSET; i<argc; ++i) {
+	strcpy(levp[0], ARGS_NAME);
+	size = ARGS_OFFSET;
+	for (i=0; i<argc; ++i) {
+	    if (i)
+		levp[0][size++] = ' ';
 	    strcpy(levp[0]+size, argv[i]);
 	    size += strlen(argv[i]);
-	    levp[0][size++] = ' ';
 	}
-	levp[0][--size] = '\0';                    /* <NUL> terminate */
 
 	/* the "SERVICE" variable */
 
-#define SERVICE_OFFSET    8                    /*  strlen('SERVICE=');  */
 #define SERVICE_NAME      "SERVICE="
+#define SERVICE_OFFSET    (sizeof(SERVICE_NAME) - 1)
 
 	retval = pam_get_item(pamh, PAM_SERVICE, &tmp);
 	if (retval != PAM_SUCCESS || tmp == NULL) {
@@ -168,17 +162,16 @@ static int process_args(pam_handle_t *pamh
 	    return -1;
 	}
 
-	strncpy(levp[1],SERVICE_NAME,SERVICE_OFFSET);
+	strcpy(levp[1], SERVICE_NAME);
 	strcpy(levp[1]+SERVICE_OFFSET, tmp);
 	levp[1][size] = '\0';                      /* <NUL> terminate */
 
 	/* the "USER" variable */
 
-#define USER_OFFSET    5                          /*  strlen('USER=');  */
 #define USER_NAME      "USER="
+#define USER_OFFSET    (sizeof(USER_NAME) - 1)
 
-	pam_get_user(pamh, &user, NULL);
-	if (user == NULL) {
+	if (pam_get_user(pamh, &user, NULL) != PAM_SUCCESS) {
 	    user = "<unknown>";
 	}
 	size = USER_OFFSET+strlen(user);
@@ -194,14 +187,14 @@ static int process_args(pam_handle_t *pamh
 	    return -1;
 	}
 
-	strncpy(levp[2],USER_NAME,USER_OFFSET);
+	strcpy(levp[2], USER_NAME);
 	strcpy(levp[2]+USER_OFFSET, user);
 	levp[2][size] = '\0';                      /* <NUL> terminate */
 
 	/* the "USER" variable */
 
-#define TYPE_OFFSET    5                          /*  strlen('TYPE=');  */
 #define TYPE_NAME      "TYPE="
+#define TYPE_OFFSET    (sizeof(TYPE_NAME) - 1)
 
 	size = TYPE_OFFSET+strlen(type);
 
@@ -217,7 +210,7 @@ static int process_args(pam_handle_t *pamh
 	    return -1;
 	}
 
-	strncpy(levp[3],TYPE_NAME,TYPE_OFFSET);
+	strcpy(levp[3], TYPE_NAME);
 	strcpy(levp[3]+TYPE_OFFSET, type);
 	levp[3][size] = '\0';                      /* <NUL> terminate */
 
@@ -253,7 +246,7 @@ static void free_evp(char *evp[])
 
 static int
 set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
-	    const char **evp, const char *filtername)
+	    char * const evp[], const char *filtername)
 {
     int status=-1;
     char* terminal = NULL;
@@ -261,7 +254,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
     int fd[2], child=0, child2=0, aterminal;
 
     if (filtername == NULL || *filtername != '/') {
-	pam_syslog(pamh, LOG_ALERT,
+	pam_syslog(pamh, LOG_ERR,
 		   "filtername not permitted; full pathname required");
 	return PAM_ABORT;
     }
@@ -296,7 +289,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	    struct termios t_mode = stored_mode;
 
 	    t_mode.c_iflag = 0;            /* no input control */
-	    t_mode.c_oflag &= ~OPOST;      /* no ouput post processing */
+	    t_mode.c_oflag &= ~OPOST;      /* no output post processing */
 
 	    /* no signals, canonical input, echoing, upper/lower output */
 #ifdef XCASE
@@ -310,7 +303,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	    t_mode.c_cc[VTIME] = 0;          /* 0/10th second for chars */
 
 	    if ( tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_mode) < 0 ) {
-		pam_syslog(pamh, LOG_WARNING,
+		pam_syslog(pamh, LOG_ERR,
 			   "couldn't put terminal in RAW mode: %m");
 		close(fd[0]);
 		return PAM_ABORT;
@@ -329,7 +322,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	 */
 
 	if ( socketpair(AF_UNIX, SOCK_STREAM, 0, fd) < 0 ) {
-	    pam_syslog(pamh, LOG_CRIT, "couldn't open a stream pipe: %m");
+	    pam_syslog(pamh, LOG_ERR, "couldn't open a stream pipe: %m");
 	    return PAM_ABORT;
 	}
     }
@@ -338,9 +331,14 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 
     if ( (child = fork()) < 0 ) {
 
-	pam_syslog(pamh, LOG_WARNING, "first fork failed: %m");
+	pam_syslog(pamh, LOG_ERR, "first fork failed: %m");
 	if (aterminal) {
 		(void) tcsetattr(STDIN_FILENO, TCSAFLUSH, &stored_mode);
+		close(fd[0]);
+	} else {
+		/* Socket pair */
+		close(fd[0]);
+		close(fd[1]);
 	}
 
 	return PAM_AUTH_ERR;
@@ -356,7 +354,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	    int t = open("/dev/tty", O_RDWR|O_NOCTTY);
 #else
 	    int t = open("/dev/tty",O_RDWR);
-	    if (t > 0) {
+	    if (t >= 0) {
 		(void) ioctl(t, TIOCNOTTY, NULL);
 		close(t);
 	    }
@@ -364,20 +362,20 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 
 	    /* make this process it's own process leader */
 	    if (setsid() == -1) {
-		pam_syslog(pamh, LOG_WARNING,
+		pam_syslog(pamh, LOG_ERR,
 			   "child cannot become new session: %m");
 		return PAM_ABORT;
 	    }
 
 	    /* grant slave terminal */
 	    if (grantpt (fd[0]) < 0) {
-		pam_syslog(pamh, LOG_WARNING, "Cannot grant acccess to slave terminal");
+		pam_syslog(pamh, LOG_ERR, "Cannot grant access to slave terminal");
 		return PAM_ABORT;
 	    }
 
 	    /* unlock slave terminal */
 	    if (unlockpt (fd[0]) < 0) {
-		pam_syslog(pamh, LOG_WARNING, "Cannot unlock slave terminal");
+		pam_syslog(pamh, LOG_ERR, "Cannot unlock slave terminal");
 		return PAM_ABORT;
 	    }
 
@@ -385,7 +383,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	    terminal = ptsname(fd[0]); /* returned value should not be freed */
 
 	    if (terminal == NULL) {
-		pam_syslog(pamh, LOG_WARNING,
+		pam_syslog(pamh, LOG_ERR,
 			   "Cannot get the name of the slave terminal: %m");
 		return PAM_ABORT;
 	    }
@@ -394,7 +392,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	    close(fd[0]);      /* process is the child -- uses line fd[1] */
 
 	    if (fd[1] < 0) {
-		pam_syslog(pamh, LOG_WARNING,
+		pam_syslog(pamh, LOG_ERR,
 			   "cannot open slave terminal: %s: %m", terminal);
 		return PAM_ABORT;
 	    }
@@ -403,7 +401,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	       parent's was before we set it into RAW mode */
 
 	    if ( tcsetattr(fd[1], TCSANOW, &stored_mode) < 0 ) {
-		pam_syslog(pamh, LOG_WARNING,
+		pam_syslog(pamh, LOG_ERR,
 			   "cannot set slave terminal mode: %s: %m", terminal);
 		close(fd[1]);
 		return PAM_ABORT;
@@ -419,7 +417,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	if ( dup2(fd[1],STDIN_FILENO) != STDIN_FILENO ||
 	     dup2(fd[1],STDOUT_FILENO) != STDOUT_FILENO ||
 	     dup2(fd[1],STDERR_FILENO) != STDERR_FILENO )  {
-	    pam_syslog(pamh, LOG_WARNING,
+	    pam_syslog(pamh, LOG_ERR,
 		       "unable to re-assign STDIN/OUT/ERR: %m");
 	    close(fd[1]);
 	    return PAM_ABORT;
@@ -430,7 +428,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	if ( fcntl(STDIN_FILENO, F_SETFD, 0) ||
 	     fcntl(STDOUT_FILENO,F_SETFD, 0) ||
 	     fcntl(STDERR_FILENO,F_SETFD, 0) ) {
-	    pam_syslog(pamh, LOG_WARNING,
+	    pam_syslog(pamh, LOG_ERR,
 		       "unable to re-assign STDIN/OUT/ERR: %m");
 	    return PAM_ABORT;
 	}
@@ -439,7 +437,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 
 	close(fd[1]);
 
-	/* the current process is now aparently working with filtered
+	/* the current process is now apparently working with filtered
 	   stdio/stdout/stderr --- success! */
 
 	return PAM_SUCCESS;
@@ -457,7 +455,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 
     if ( (child2 = fork()) < 0 ) {
 
-	pam_syslog(pamh, LOG_WARNING, "filter fork failed: %m");
+	pam_syslog(pamh, LOG_ERR, "filter fork failed: %m");
 	child2 = 0;
 
     } else if ( child2 == 0 ) {              /* exec the child filter */
@@ -465,7 +463,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	if ( dup2(fd[0],APPIN_FILENO) != APPIN_FILENO ||
 	     dup2(fd[0],APPOUT_FILENO) != APPOUT_FILENO ||
 	     dup2(fd[0],APPERR_FILENO) != APPERR_FILENO )  {
-	    pam_syslog(pamh, LOG_WARNING,
+	    pam_syslog(pamh, LOG_ERR,
 		       "unable to re-assign APPIN/OUT/ERR: %m");
 	    close(fd[0]);
 	    _exit(1);
@@ -476,7 +474,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 	if ( fcntl(APPIN_FILENO, F_SETFD, 0) == -1 ||
 	     fcntl(APPOUT_FILENO,F_SETFD, 0) == -1 ||
 	     fcntl(APPERR_FILENO,F_SETFD, 0) == -1 ) {
-	    pam_syslog(pamh, LOG_WARNING,
+	    pam_syslog(pamh, LOG_ERR,
 		       "unable to retain APPIN/OUT/ERR: %m");
 	    close(APPIN_FILENO);
 	    close(APPOUT_FILENO);
@@ -490,7 +488,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 
 	/* getting to here is an error */
 
-	pam_syslog(pamh, LOG_ALERT, "filter: %s: %m", filtername);
+	pam_syslog(pamh, LOG_ERR, "filter: %s: %m", filtername);
 	_exit(1);
 
     } else {           /* wait for either of the two children to exit */
@@ -519,7 +517,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 		    child2 = 0;
 	    } else {
 
-		pam_syslog(pamh, LOG_ALERT,
+		pam_syslog(pamh, LOG_ERR,
 			   "programming error <chid=%d,lstatus=%x> "
 			   "in file %s at line %d",
 			   chid, lstatus, __FILE__, __LINE__);
@@ -557,7 +555,7 @@ set_filter (pam_handle_t *pamh, int flags UNUSED, int ctrl,
 
 	} else {
 
-	    pam_syslog(pamh, LOG_ALERT,
+	    pam_syslog(pamh, LOG_ERR,
 		       "programming error <chid=%d,lstatus=%x> "
 		       "in file %s at line %d",
 		       chid, lstatus, __FILE__, __LINE__);
@@ -627,8 +625,7 @@ static int need_a_filter(pam_handle_t *pamh
     }
 
     if (retval == PAM_SUCCESS && (ctrl & which_run)) {
-	retval = set_filter(pamh, flags, ctrl
-			    , (const char **)evp, filterfile);
+	retval = set_filter(pamh, flags, ctrl, evp, filterfile);
     }
 
     if (retval == PAM_SUCCESS
@@ -658,23 +655,23 @@ static int need_a_filter(pam_handle_t *pamh
 
 /* ------------------ authentication ----------------- */
 
-PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh
-				   , int flags, int argc, const char **argv)
+int pam_sm_authenticate(pam_handle_t *pamh,
+			int flags, int argc, const char **argv)
 {
     return need_a_filter(pamh, flags, argc, argv
 			 , "authenticate", FILTER_RUN1);
 }
 
-PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags
-			      , int argc, const char **argv)
+int pam_sm_setcred(pam_handle_t *pamh, int flags,
+		   int argc, const char **argv)
 {
     return need_a_filter(pamh, flags, argc, argv, "setcred", FILTER_RUN2);
 }
 
 /* --------------- account management ---------------- */
 
-PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc,
-                              const char **argv)
+int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc,
+                     const char **argv)
 {
     return need_a_filter(pamh, flags, argc, argv
 			 , "setcred", FILTER_RUN1|FILTER_RUN2 );
@@ -682,15 +679,15 @@ PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc,
 
 /* --------------- session management ---------------- */
 
-PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags
-				   , int argc, const char **argv)
+int pam_sm_open_session(pam_handle_t *pamh, int flags,
+			int argc, const char **argv)
 {
     return need_a_filter(pamh, flags, argc, argv
 			 , "open_session", FILTER_RUN1);
 }
 
-PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags
-			      , int argc, const char **argv)
+int pam_sm_close_session(pam_handle_t *pamh, int flags,
+                         int argc, const char **argv)
 {
     return need_a_filter(pamh, flags, argc, argv
 			 , "close_session", FILTER_RUN2);
@@ -699,8 +696,8 @@ PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags
 /* --------- updating authentication tokens --------- */
 
 
-PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags
-				, int argc, const char **argv)
+int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
+		     int argc, const char **argv)
 {
     int runN;
 
@@ -715,19 +712,3 @@ PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags
 
     return need_a_filter(pamh, flags, argc, argv, "chauthtok", runN);
 }
-
-#ifdef PAM_STATIC
-
-/* ------------ stuff for static modules ------------ */
-
-struct pam_module _pam_filter_modstruct = {
-    "pam_filter",
-    pam_sm_authenticate,
-    pam_sm_setcred,
-    pam_sm_acct_mgmt,
-    pam_sm_open_session,
-    pam_sm_close_session,
-    pam_sm_chauthtok,
-};
-
-#endif
