@@ -1,4 +1,6 @@
 /*
+ * pam_xauth module
+ *
  * Copyright 2001-2003 Red Hat, Inc.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -50,9 +52,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
-#include <unistd.h>
-
-#define PAM_SM_SESSION
 
 #include <security/pam_modules.h>
 #include <security/_pam_macros.h>
@@ -62,9 +61,11 @@
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
 #include <selinux/label.h>
-#include <sys/stat.h>
 #endif
 
+#include "pam_cc_compat.h"
+#include "pam_inline.h"
+
 #define DATANAME "pam_xauth_cookie_file"
 #define XAUTHENV "XAUTHORITY"
 #define HOMEENV  "HOME"
@@ -103,9 +104,11 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 
 	/* Create stdio pipery. */
 	if (pipe(ipipe) == -1) {
+		pam_syslog(pamh, LOG_ERR, "Could not create pipe: %m");
 		return -1;
 	}
 	if (pipe(opipe) == -1) {
+		pam_syslog(pamh, LOG_ERR, "Could not create pipe: %m");
 		close(ipipe[0]);
 		close(ipipe[1]);
 		return -1;
@@ -114,6 +117,7 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 	/* Fork off a child. */
 	child = fork();
 	if (child == -1) {
+		pam_syslog(pamh, LOG_ERR, "Could not fork: %m");
 		close(ipipe[0]);
 		close(ipipe[1]);
 		close(opipe[0]);
@@ -124,9 +128,7 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 	if (child == 0) {
 		/* We're the child. */
 		size_t j;
-		char *args[10];
-		const char *tmp;
-		int maxopened;
+		const char *args[10];
 		/* Drop privileges. */
 		if (setgid(gid) == -1)
 		  {
@@ -148,42 +150,50 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 				(unsigned long) geteuid ());
 		    _exit (err);
 		  }
-		/* Initialize the argument list. */
-		memset(args, 0, sizeof(args));
 		/* Set the pipe descriptors up as stdin and stdout, and close
 		 * everything else, including the original values for the
 		 * descriptors. */
-		dup2(ipipe[0], STDIN_FILENO);
-		dup2(opipe[1], STDOUT_FILENO);
-		maxopened = (int)sysconf(_SC_OPEN_MAX);
-		for (i = 0; i < maxopened; i++) {
-			if ((i != STDIN_FILENO) && (i != STDOUT_FILENO)) {
-				close(i);
-			}
+		if (dup2(ipipe[0], STDIN_FILENO) != STDIN_FILENO) {
+		    int err = errno;
+		    pam_syslog(pamh, LOG_ERR, "dup2 of %s failed: %m", "stdin");
+		    _exit(err);
 		}
+		if (dup2(opipe[1], STDOUT_FILENO) != STDOUT_FILENO) {
+		    int err = errno;
+		    pam_syslog(pamh, LOG_ERR, "dup2 of %s failed: %m", "stdout");
+		    _exit(err);
+		}
+		if (pam_modutil_sanitize_helper_fds(pamh, PAM_MODUTIL_IGNORE_FD,
+						    PAM_MODUTIL_IGNORE_FD,
+						    PAM_MODUTIL_NULL_FD) < 0) {
+		    _exit(1);
+		}
+		/* Initialize the argument list. */
+		memset(args, 0, sizeof(args));
 		/* Convert the varargs list into a regular array of strings. */
 		va_start(ap, command);
-		args[0] = strdup(command);
-		for (j = 1; j < ((sizeof(args) / sizeof(args[0])) - 1); j++) {
-			tmp = va_arg(ap, const char*);
-			if (tmp == NULL) {
+		args[0] = command;
+		for (j = 1; j < PAM_ARRAY_SIZE(args) - 1; j++) {
+			args[j] = va_arg(ap, const char*);
+			if (args[j] == NULL) {
 				break;
 			}
-			args[j] = strdup(tmp);
 		}
 		/* Run the command. */
-		execv(command, args);
+		DIAG_PUSH_IGNORE_CAST_QUAL;
+		execv(command, (char *const *) args);
+		DIAG_POP_IGNORE_CAST_QUAL;
 		/* Never reached. */
 		_exit(1);
 	}
 
 	/* We're the parent, so close the other ends of the pipes. */
-	close(ipipe[0]);
 	close(opipe[1]);
 	/* Send input to the process (if we have any), then send an EOF. */
 	if (input) {
 		(void)pam_modutil_write(ipipe[1], input, strlen(input));
 	}
+	close(ipipe[0]); /* close here to avoid possible SIGPIPE above */
 	close(ipipe[1]);
 
 	/* Read data output until we run out of stuff to read. */
@@ -354,17 +364,19 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 	/* Parse arguments.  We don't understand many, so no sense in breaking
 	 * this into a separate function. */
 	for (i = 0; i < argc; i++) {
+		const char *str;
+
 		if (strcmp(argv[i], "debug") == 0) {
 			debug = 1;
 			continue;
 		}
-		if (strncmp(argv[i], "xauthpath=", 10) == 0) {
-			xauth = argv[i] + 10;
+		if ((str = pam_str_skip_prefix(argv[i], "xauthpath=")) != NULL) {
+			xauth = str;
 			continue;
 		}
-		if (strncmp(argv[i], "targetuser=", 11) == 0) {
-			long l = strtol(argv[i] + 11, &tmp, 10);
-			if ((strlen(argv[i] + 11) > 0) && (*tmp == '\0')) {
+		if ((str = pam_str_skip_prefix(argv[i], "targetuser=")) != NULL) {
+			long l = strtol(str, &tmp, 10);
+			if ((*str != '\0') && (*tmp == '\0')) {
 				targetuser = l;
 			} else {
 				pam_syslog(pamh, LOG_WARNING,
@@ -373,9 +385,9 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 			}
 			continue;
 		}
-		if (strncmp(argv[i], "systemuser=", 11) == 0) {
-			long l = strtol(argv[i] + 11, &tmp, 10);
-			if ((strlen(argv[i] + 11) > 0) && (*tmp == '\0')) {
+		if ((str = pam_str_skip_prefix(argv[i], "systemuser=")) != NULL) {
+			long l = strtol(str, &tmp, 10);
+			if ((*str != '\0') && (*tmp == '\0')) {
 				systemuser = l;
 			} else {
 				pam_syslog(pamh, LOG_WARNING,
@@ -390,7 +402,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 
 	if (xauth == NULL) {
 	        size_t j;
-		for (j = 0; j < sizeof(xauthpaths)/sizeof(xauthpaths[0]); j++) {
+		for (j = 0; j < PAM_ARRAY_SIZE(xauthpaths); j++) {
 			if (access(xauthpaths[j], X_OK) == 0) {
 				xauth = xauthpaths[j];
 				break;
@@ -413,8 +425,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 
 	/* Read the target user's name. */
 	if (pam_get_user(pamh, &user, NULL) != PAM_SUCCESS) {
-		pam_syslog(pamh, LOG_ERR,
-			   "error determining target user's name");
+		pam_syslog(pamh, LOG_NOTICE, "cannot determine user name");
 		retval = PAM_SESSION_ERR;
 		goto cleanup;
 	}
@@ -430,7 +441,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 	 * on the xauthority file we create later on. */
 	tpwd = pam_modutil_getpwnam(pamh, user);
 	if (tpwd == NULL) {
-		pam_syslog(pamh, LOG_ERR,
+		pam_syslog(pamh, LOG_NOTICE,
 			   "error determining target user's UID");
 		retval = PAM_SESSION_ERR;
 		goto cleanup;
@@ -459,24 +470,33 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 		goto cleanup;
 	}
 
-	/* Check that both users are amenable to this.  By default, this
-	 * boils down to this policy:
-	 * export(ruser=root): only if <user> is listed in .xauth/export
-	 * export(ruser=*) if <user> is listed in .xauth/export, or
-	 *                 if .xauth/export does not exist
-	 * import(user=*): if <ruser> is listed in .xauth/import, or
-	 *                 if .xauth/import does not exist */
-	i = (getuid() != 0 || tpwd->pw_uid == 0) ? PAM_SUCCESS : PAM_PERM_DENIED;
-	i = check_acl(pamh, "export", rpwd->pw_name, user, i, debug);
-	if (i != PAM_SUCCESS) {
-		retval = PAM_SESSION_ERR;
-		goto cleanup;
-	}
-	i = PAM_SUCCESS;
-	i = check_acl(pamh, "import", user, rpwd->pw_name, i, debug);
-	if (i != PAM_SUCCESS) {
-		retval = PAM_SESSION_ERR;
-		goto cleanup;
+
+	/* If current user and the target user are the same, don't
+	   check the ACL list, but forward X11 */
+	if (strcmp (rpwd->pw_name, tpwd->pw_name) != 0) {
+
+	  /* Check that both users are amenable to this.  By default, this
+	   * boils down to this policy:
+	   * export(ruser=root): only if <user> is listed in .xauth/export
+	   * export(ruser=*) if <user> is listed in .xauth/export, or
+	   *                 if .xauth/export does not exist
+	   * import(user=*): if <ruser> is listed in .xauth/import, or
+	   *                 if .xauth/import does not exist */
+	  i = (getuid() != 0 || tpwd->pw_uid == 0) ? PAM_SUCCESS : PAM_PERM_DENIED;
+	  i = check_acl(pamh, "export", rpwd->pw_name, user, i, debug);
+	  if (i != PAM_SUCCESS) {
+	    retval = PAM_SESSION_ERR;
+	    goto cleanup;
+	  }
+	  i = PAM_SUCCESS;
+	  i = check_acl(pamh, "import", user, rpwd->pw_name, i, debug);
+	  if (i != PAM_SUCCESS) {
+	    retval = PAM_SESSION_ERR;
+	    goto cleanup;
+	  }
+	}  else {
+	  if (debug)
+	    pam_syslog (pamh, LOG_DEBUG, "current and target user are the same, forward X11");
 	}
 
 	/* Figure out where the source user's .Xauthority file is. */
@@ -512,14 +532,14 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 			  xauth, "-f", cookiefile, "nlist", display,
 			  NULL) == 0) {
 #ifdef WITH_SELINUX
-		security_context_t context = NULL;
+		char *context_raw = NULL;
 #endif
 		PAM_MODUTIL_DEF_PRIVS(privs);
 
 		/* Check that we got a cookie.  If not, we get creative. */
 		if (((cookie == NULL) || (strlen(cookie) == 0)) &&
-		    ((strncmp(display, "localhost:", 10) == 0) ||
-		     (strncmp(display, "localhost/unix:", 15) == 0))) {
+		    (pam_str_skip_prefix(display, "localhost:") != NULL ||
+		     pam_str_skip_prefix(display, "localhost/unix:") != NULL)) {
 			char *t, *screen;
 			size_t tlen, slen;
 			/* Free the useless cookie string. */
@@ -606,16 +626,16 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 		if (is_selinux_enabled() > 0) {
 			struct selabel_handle *ctx = selabel_open(SELABEL_CTX_FILE, NULL, 0);
 			if (ctx != NULL) {
-				if (selabel_lookup(ctx, &context,
-						   xauthority + sizeof(XAUTHENV), S_IFREG) != 0) {
+				if (selabel_lookup_raw(ctx, &context_raw,
+						       xauthority + sizeof(XAUTHENV), S_IFREG) != 0) {
 					pam_syslog(pamh, LOG_WARNING,
 						   "could not get SELinux label for '%s'",
 						   xauthority + sizeof(XAUTHENV));
 				}
 				selabel_close(ctx);
-				if (setfscreatecon(context)) {
+				if (setfscreatecon_raw(context_raw)) {
 					pam_syslog(pamh, LOG_WARNING,
-						   "setfscreatecon(%s) failed: %m", context);
+						   "setfscreatecon_raw(%s) failed: %m", context_raw);
 				}
 			}
 		}
@@ -626,9 +646,9 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 				   "error creating temporary file `%s': %m",
 				   xauthority + sizeof(XAUTHENV));
 #ifdef WITH_SELINUX
-		if (context != NULL) {
-			free(context);
-			setfscreatecon(NULL);
+		if (context_raw != NULL) {
+			free(context_raw);
+			setfscreatecon_raw(NULL);
 		}
 #endif /* WITH_SELINUX */
 		if (fd >= 0)
@@ -667,7 +687,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 
 		  if (asprintf(&d, "DISPLAY=%s", display) < 0)
 		    {
-		      pam_syslog(pamh, LOG_ERR, "out of memory");
+		      pam_syslog(pamh, LOG_CRIT, "out of memory");
 		      cookiefile = NULL;
 		      retval = PAM_SESSION_ERR;
 		      goto cleanup;
@@ -684,7 +704,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 		  char *d;
 
 		  if (asprintf(&d, "XAUTHLOCALHOSTNAME=%s", xauthlocalhostname) < 0) {
-		    pam_syslog(pamh, LOG_ERR, "out of memory");
+		    pam_syslog(pamh, LOG_CRIT, "out of memory");
 		    retval = PAM_SESSION_ERR;
 		    goto cleanup;
 		  }
@@ -750,23 +770,22 @@ pam_sm_close_session (pam_handle_t *pamh, int flags UNUSED,
 			debug = 1;
 			continue;
 		}
-		if (strncmp(argv[i], "xauthpath=", 10) == 0)
+		if (pam_str_skip_prefix(argv[i], "xauthpath=") != NULL)
 			continue;
-		if (strncmp(argv[i], "systemuser=", 11) == 0)
+		if (pam_str_skip_prefix(argv[i], "systemuser=") != NULL)
 			continue;
-		if (strncmp(argv[i], "targetuser=", 11) == 0)
+		if (pam_str_skip_prefix(argv[i], "targetuser=") != NULL)
 			continue;
 		pam_syslog(pamh, LOG_WARNING, "unrecognized option `%s'",
 		       argv[i]);
 	}
 
 	if (pam_get_user(pamh, &user, NULL) != PAM_SUCCESS) {
-		pam_syslog(pamh, LOG_ERR,
-			   "error determining target user's name");
+		pam_syslog(pamh, LOG_NOTICE, "cannot determine user name");
 		return PAM_SESSION_ERR;
 	}
 	if (!(tpwd = pam_modutil_getpwnam(pamh, user))) {
-		pam_syslog(pamh, LOG_ERR,
+		pam_syslog(pamh, LOG_NOTICE,
 			   "error determining target user's UID");
 		return PAM_SESSION_ERR;
 	}
@@ -782,16 +801,3 @@ pam_sm_close_session (pam_handle_t *pamh, int flags UNUSED,
 
 	return PAM_SUCCESS;
 }
-
-/* static module data */
-#ifdef PAM_STATIC
-struct pam_module _pam_xauth_modstruct = {
-  "pam_xauth",
-  NULL,
-  NULL,
-  NULL,
-  pam_sm_open_session,
-  pam_sm_close_session,
-  NULL
-};
-#endif
